/*
 * testAddCertif.h
 *
 * This file is part of CAMPARY Library
 *
 * Copyright (C) 2014 - 
 *
 * CAMPARY Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CAMPARY Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
 
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
 
#include <ctime>
#include <cstdio>
#include <iostream>
#include <string>
using namespace std;
#include <gmp.h>
#include <sys/time.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <mpfr.h>
#include "../src/multi_prec.h"

#define K 16
#define L 16
#define R 16

int test(int trials){
  int failCases = 0;
	int prec = (K>L)?K:L;
	prec = (R>prec)?R:prec;

	mpfr_t mpf1, mpf2, mpfRes, mpfMP, mpfErr, mpfBound1, mpfBound2;
	mpfr_init2(mpf1, prec*53 + 106);
	mpfr_init2(mpf2, prec*53 + 106);
	mpfr_init2(mpfMP, prec*53 + 106);
	mpfr_init2(mpfRes, prec*53 + 106);
	mpfr_init2(mpfErr, prec*53 + 106);
  mpfr_init2(mpfBound1, prec*53 + 106);
  mpfr_init2(mpfBound2, prec*53 + 106); 

  mpfr_set_d(mpfBound1, pow(2,-52*R), GMP_RNDN);

	multi_prec<K> mp1;
	multi_prec<L> mp2;
	multi_prec<R> mpRes;

	for (int j=1; j<=trials; j++){
		/********* MPrec computations *********/
		mp1.randomInit_ulp();
		mp2.randomInit_ulp();

		certifAddExpans(mpRes, mp1, mp2);

		/********* MPFR computations *********/
	  mpfr_set_d(mpf1, mp1[0], GMP_RNDN);
	  for(int i=1; i<K; i++)
 	  	mpfr_add_d(mpf1, mpf1, mp1[i], GMP_RNDN);
	  mpfr_set_d(mpf2, mp2[0], GMP_RNDN);
	  for(int i=1; i<L; i++)
	  	mpfr_add_d(mpf2, mpf2, mp2[i], GMP_RNDN);

	  mpfr_add(mpfRes, mpf1, mpf2, GMP_RNDN);

		/********* Error computation *********/
	  mpfr_set_d(mpfMP, mpRes[0] , GMP_RNDN);
	  for(int i=1; i<R; i++)
	  	mpfr_add_d(mpfMP, mpfMP, mpRes[i], GMP_RNDN);
		mpfr_sub(mpfErr, mpfRes, mpfMP, GMP_RNDN);

		mpfr_abs(mpfErr, mpfErr, GMP_RNDN);
    mpfr_mul_d(mpfBound2, mpfBound1, fabs(mpRes[0]), GMP_RNDN);

	  if( (mpfr_cmp(mpfErr, mpfBound2) <= 0) || (mpfr_cmp_d(mpfErr, 0.0) == 0) ){}
	  else{
			cout << "------- Test " << j << " ---------" << endl;

			cout << "------- MPrec Input ---------" << endl;
    	mp1.prettyPrintBin();
    	mp2.prettyPrintBin();

			cout << "------- MPrec Output ---------" << endl;
    	mpRes.prettyPrintBin();

			cout << "------- MPFR Input ---------" << endl;
    	mpfr_out_str(stdout, 2, 0, mpf1, GMP_RNDN);
    	cout << endl;
    	mpfr_out_str(stdout, 2, 0, mpf2, GMP_RNDN);
    	cout << endl;

			cout << "------- MPFR Output ---------" << endl;
    	mpfr_out_str(stdout, 2, 0, mpfRes, GMP_RNDN);	    
			cout << endl;

	    cout << "Fail, error = ";
	    mpfr_out_str(stdout, 2, 0, mpfErr, GMP_RNDN);
	    cout << endl;

      cout << "MPFR bound1 = ";
      mpfr_out_str(stdout, 2, 0, mpfBound1, GMP_RNDN);
      cout << endl;

      cout << "MPFR bound2 = ";
      mpfr_out_str(stdout, 2, 0, mpfBound2, GMP_RNDN);
      cout << endl;

			failCases++;
	  }
	}

  mpfr_clear(mpfBound1);
  mpfr_clear(mpfBound2);
	mpfr_clear(mpf1);
	mpfr_clear(mpf2);
	mpfr_clear(mpfMP);
	mpfr_clear(mpfRes);
	mpfr_clear(mpfErr);
  return failCases;
}

int main(int argc, char *argv[]) {
	if(argc<2){
		cout << "Usage: ./testAddCertif trials" << endl; 
   	return 0;
  }

	int trials = atoi(argv[1]);

  cout << "Testing " << trials << " MPrec 'certified' additions with: " << K << " doubles + " << L << " doubles = " << R << " doubles " << endl;
  int failCases = test(trials);
  cout << "Total failed cases:" << failCases << endl;
  return 0;
}
