/*
 * multi_prec.h
 *
 * This file is part of CAMPARY Library
 *
 * Copyright (C) 2014 - 
 *
 * CAMPARY Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CAMPARY Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
 
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */

#ifndef _multi_prec_h
#define _multi_prec_h

#include <stdio.h>
#include <math.h>
#include <fenv.h>
#include <cstdio>
#include <string.h>
#include <iostream>
#include <stdlib.h>
using namespace std;

#include "helper_fct.h"
#include "errorFreeTransf.h"
#include "renorm.h"
#include "Specs/specRenorm.h"
#include "addition.h"
#include "Specs/specAddition.h"
#include "multiplication.h"
#include "Specs/specMultiplication.h"
#include "newton.h"
#include "Specs/specNewton.h"

/**forward declarations**/
template <int prec> class multi_prec ;

/**template friends**/
template <int prec>
multi_prec<prec> abs(const multi_prec<prec> &mp);
template <int prec>
multi_prec<prec> sqrt(const multi_prec<prec> &mp);
template <int prec>
multi_prec<prec> invSqrt(const multi_prec<prec> &mp);

template <int prec>
void renorm(multi_prec<prec> &mp);
template <int prec>
void renorm_rand(multi_prec<prec> &mp);
template <int prec>
void renorm_2ndL(multi_prec<prec> &mp);

//********************* ADD *********************//
template <int pR, int p1, int p2>
void    certifAddExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
template <int pR, int p1>
void  certifAddExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
template <int pR, int p1, int p2>
void     truncAddExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
template <int pR, int p1>
void   truncAddExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
template <int pR, int p1, int p2>
void   QD_LikeAddExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
template <int pR, int p1>
void QD_LikeAddExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );

//********************* MUL *********************//
template <int pR, int p1, int p2>
void    certifMulExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
template <int pR, int p1>
void  certifMulExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
template <int pR, int p1, int p2>
void     truncMulExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
template <int pR, int p1>
void   truncMulExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
template <int pR, int p1, int p2>
void   QD_LikeMulExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
template <int pR, int p1>
void QD_LikeMulExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );

//********************* DIV *********************//
template <int pR, int p1>
void   invExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
template <int pR>
void invExpans_d( multi_prec<pR> &res, const float val );
template <int pR, int p1, int p2>
void   divExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
template <int pR, int p1>
void divExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
template <int pR, int p1>
void divExpans_d( multi_prec<pR> &res, const float val, const multi_prec<p1> &mp1 );

//********************* SQRT *********************//
template <int pR, int p1>
void      invSqrtExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
template <int pR>
void    invSqrtExpans_d( multi_prec<pR> &res, const float val );
template <int pR, int p1>
void   sqrtNewtonExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
template <int pR>
void sqrtNewtonExpans_d( multi_prec<pR> &res, const float val );
template <int pR, int p1>
void    sqrtHeronExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
template <int pR>
void  sqrtHeronExpans_d( multi_prec<pR> &res, const float val );

template <int prec>
class multi_prec{
private:
	float data[prec];
public:
	//---------constructors--------------
	multi_prec(){};
	multi_prec(const float x);
	multi_prec(const float *datap, const int precp);
	multi_prec(const char *s);
	template<int precP>	multi_prec(const multi_prec<precP> *mp);
	multi_prec(const multi_prec<prec> *mp);

	//----------geters & setters--------------
	int getPrec() const;
	const float* getData() const;
	void setData(const float *datap, const int precp);
	void setData(const float *datap);
	void setElement(const float datap, const int index);

 	/**pretty print**/
 	void prettyPrint(){
  	printf("Prec = %d\n", prec); 
  	for(int i=0; i<prec; i++) printf("   Data[%d] = %e\n",i,data[i]);
 	}
 	void prettyPrintBin(){
  	printf("Prec = %d\n", prec); 
  	for(int i=0; i<prec; i++)	printf("   Data[%d] = %a;\n",i,data[i]);
 	}
  void prettyPrintBin_UnevalSum(){
  	for(int i=0; i<prec-1; i++) printf("%a + ", data[i]);
  	printf("%a;", data[prec-1]);
  }
  
  char* prettyPrintBF(){
		size_t needed = snprintf(NULL, 0, "%e", data[0]);
		if (prec>1)
			for(int i=1; i<prec; i++)	needed += snprintf(NULL, 0, "%e ", data[i]);
    
		char *ch;
		ch =(char *) malloc((2*needed++)*sizeof(char));
     
		sprintf(ch, "%e ", data[0]);
		if (prec>1)
			for(int i=1; i<prec; i++) sprintf(&ch[strlen(ch)], "%e ", data[i]);

		sprintf(&ch[strlen(ch)], "%c", '\0');
		return ch;
 	}

	//----------operators--------------
	/** operator [] overloading  **/
	float operator [](const int i)const {return data[i];}

	/** Puts the value other in data[0] of the current multi_prec variable  **/
	multi_prec& operator =(const float other){
		setData(&other, 1);
		return *this;
	}
	/** Transfers the multi_prec parameter other in the current multi_prec variable **/
	multi_prec<prec>& operator =(const multi_prec<prec>& other){
		if(this != &other) setData(other.getData(), prec); /* check against self assingment */
		return *this;
	}
	template <int pS>
	multi_prec& operator =(const multi_prec<pS>& other){
	 	setData(other.getData(),pS);
	 	return *this;
	}
	multi_prec& operator =(const char *s){
		if (read(s, *this)){
			printf("(qd_real::operator=): INPUT ERROR.");
    	*this = 0.0;
  	}
  	return *this;
	}

	/** Equality overloading**/
	template <int pS> bool operator ==(const multi_prec<pS> &mp2) const;
	bool operator ==(const multi_prec<prec> &mp2) const;
	bool operator ==(const float &mp2) const;
	template <int pS> bool operator !=(const multi_prec<pS> &mp2) const;
	bool operator !=(const multi_prec<prec> &mp2) const;
	bool operator !=(const float &mp2) const;
	template <int pS> bool operator < (const multi_prec<pS> &mp2) const;
	bool operator < (const multi_prec<prec> &mp2) const;
	bool operator < (const float &mp2) const;
	template <int pS> bool operator <=(const multi_prec<pS> &mp2) const;
	bool operator <=(const multi_prec<prec> &mp2) const;
	bool operator <=(const float &mp2) const;
	template <int pS> bool operator > (const multi_prec<pS> &mp2) const;
	bool operator > (const multi_prec<prec> &mp2) const;
	bool operator > (const float &mp2) const;
	template <int pS> bool operator >=(const multi_prec<pS> &mp2) const;
	bool operator >=(const multi_prec<prec> &mp2) const;
	bool operator >=(const float &mp2) const;

	/**Random expansion generation **/
	void randomInit_P(){ genExpans_P<prec>(data);}
  void randomInit_ulp(){ genExpans_ulp<prec>(data);}
  void randomInit_ulp(int order){ genExpans_ulp<prec>(data, order);}
  void randomInit_ulp(float first){ genExpans_ulp<prec>(data, first);}

	/**Unary - overloading**/
	friend multi_prec operator -(const multi_prec<prec> &mp1){
		multi_prec<prec> res;
		//res = multi_prec<prec>(0.);
		for(int i=0; i<prec; i++) res.data[i] = -mp1.getData()[i];
		return res;
	}
/////////////////////////////////////////////////////////////////////////
	/** operator += overloading  **/
	template <int pS>
	multi_prec<prec>& operator +=(const multi_prec<pS> &mp){
		QD_LikeAddExpans<prec,prec,pS>( *this, *this, mp );
		return *this;
	}
	multi_prec<prec>& operator +=(const float val){
		QD_LikeAddExpans_d<prec,prec>( *this, *this, val );
		return *this;
	}

  /** operator -= overloading  **/
  template <int pS>
  multi_prec<prec>& operator -=(const multi_prec<pS> &mp){
    QD_LikeAddExpans<prec,prec,pS>( *this, *this, -mp );
    return *this;
  }
  multi_prec<prec>& operator -=(const float val){
    QD_LikeAddExpans_d<prec,prec>( *this, *this, -val );
    return *this;
  }

  /** operator *= overloading  **/
  template <int pS>
  multi_prec<prec> operator *=(const multi_prec<pS> &mp){
    QD_LikeMulExpans<prec,prec,pS>( *this, *this, mp );
    return *this;
  }
  multi_prec<prec>& operator *=(const float val){
    QD_LikeMulExpans_d<prec,prec>( *this, *this, val );
    return *this;
  }

	/** operator /= overloading  **/
	template <int pS>
	multi_prec<prec>& operator /=(const multi_prec<pS> &mp){
		divExpans<prec,prec,pS>( *this, *this, mp );
		return *this;
	}
	multi_prec<prec>& operator /=(const float val){
		divExpans_d<prec,prec>( *this, *this, val );
		return *this;
	}

	/**absolute value **/
	friend multi_prec<prec> abs<>(const multi_prec<prec> &mp);
	friend multi_prec<prec> sqrt<>(const multi_prec<prec> &mp);
	friend multi_prec<prec> invSqrt<>(const multi_prec<prec> &mp);
	
	template<int pS> friend int read(const char *s, multi_prec<pS> &mp);
	
	template <int pR, int p1, int p2>
	friend multi_prec<pR> max( const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
	template <int pR, int p1, int p2>
	friend multi_prec<pR> min( const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
	
	friend void renorm<>(multi_prec<prec> &mp);
	friend void renorm_rand<>(multi_prec<prec> &mp);
	friend void renorm_2ndL<>(multi_prec<prec> &mp);

	/** operator + overloading  **/
	template <int pS>
	friend multi_prec operator +(const multi_prec<prec> &mp1, const multi_prec<pS> &mp2){
		multi_prec<(prec>pS)?prec:pS> res(mp1.getData(), prec);
		return res += mp2;
	}
	friend multi_prec operator +(const multi_prec<prec> &mp1, const float val){
		multi_prec<prec> res(mp1.getData(), prec);
		return res += val;
	}
	friend multi_prec operator +(const float val, const multi_prec<prec> &mp1){
		multi_prec<prec> res(mp1.getData(), prec);
		return res += val;
	}

	/** operator - overloading  **/
	template <int pS>
	friend multi_prec operator -(const multi_prec<prec> &mp1, const multi_prec<pS> &mp2){
		multi_prec<(prec>pS)?prec:pS> res(mp1.getData(), prec);
		return res -= mp2;
	}
	friend multi_prec operator -(const multi_prec<prec> &mp1, const float val){
		multi_prec<prec> res(mp1.getData(), prec);
		return res -= val;
	}
	friend multi_prec operator -(const float val, const multi_prec<prec> &mp1){
		multi_prec<prec> res(val);
		return res -= mp1;
	}

	/** operator * overloading  **/
	template <int pS>
	friend multi_prec operator *(const multi_prec<prec> &mp1, const multi_prec<pS> &mp2){
		multi_prec<(prec>pS)?prec:pS> res(mp1.getData(), prec);
		return res *= mp2;
	}
	friend multi_prec operator *(const multi_prec<prec> &mp1, const float val){
		multi_prec<prec> res(mp1.getData(), prec);
		return res *= val;
	}
	friend multi_prec operator *(const float val, const multi_prec<prec> &mp1){
		multi_prec<prec> res(mp1.getData(), prec);
		return res *= val;
	}

	/** operator / overloading  **/
  template <int pS>
  friend multi_prec operator /(const multi_prec<prec> &mp1, const multi_prec<pS> &mp2){
    multi_prec<(prec>pS)?prec:pS> res(mp1.getData(), prec);
    return res /= mp2;
  }
  friend multi_prec operator /(const multi_prec<prec> &mp1, const float val){
    multi_prec<prec> res(mp1.getData(), prec);
    return res /= val;
  }
  friend multi_prec operator /(const float val, const multi_prec<prec> &mp1){
    multi_prec<prec> res(val);
    return res /= mp1;
  }

	//********************* ADD *********************//
	template <int pR, int p1, int p2>
	friend void    certifAddExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
	template <int pR, int p1>
	friend void  certifAddExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
	template <int pR, int p1, int p2>
  friend void     truncAddExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
  template <int pR, int p1>
  friend void   truncAddExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
	template <int pR, int p1, int p2>
	friend void   QD_LikeAddExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
	template <int pR, int p1>
	friend void QD_LikeAddExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );

	//********************* MUL *********************//
	template <int pR, int p1, int p2>
	friend void    certifMulExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
	template <int pR, int p1>
	friend void  certifMulExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
  template <int pR, int p1, int p2>
  friend void     truncMulExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
  template <int pR, int p1>
  friend void   truncMulExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
  template <int pR, int p1, int p2>
	friend void   QD_LikeMulExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
  template <int pR, int p1>
	friend void QD_LikeMulExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );

	//********************* DIV *********************//
	template <int pR, int p1>
	friend void   invExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
	template <int pR>
	friend void invExpans_d( multi_prec<pR> &res, const float val );
	template <int pR, int p1, int p2>
	friend void   divExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 );
	template <int pR, int p1>
	friend void divExpans_d( multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val );
  template <int pR, int p1>
  friend void divExpans_d( multi_prec<pR> &res, const float val, const multi_prec<p1> &mp1 );

	//********************* SQRT *********************//
  template <int pR, int p1>
  friend void      invSqrtExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
  template <int pR>
  friend void    invSqrtExpans_d( multi_prec<pR> &res, const float val );
	template <int pR, int p1>
	friend void   sqrtNewtonExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
	template <int pR>
  friend void sqrtNewtonExpans_d( multi_prec<pR> &res, const float val );
	template <int pR, int p1>
  friend void    sqrtHeronExpans( multi_prec<pR> &res, const multi_prec<p1> &mp1 );
	template <int pR>
  friend void  sqrtHeronExpans_d( multi_prec<pR> &res, const float val );
};

/*definitions*/
//********************* Constructors *********************//
template <int prec>
multi_prec<prec>::multi_prec(const float x){
	data[0] = x;
  for(int i=1; i<prec; i++) data[i] = 0;
}
template <int prec>
multi_prec<prec>::multi_prec(const float *datap,const int precp){
  for(int i=0; i<min(prec, precp); i++)	data[i]=datap[i];
  for(int i=min(prec,precp); i<prec; i++) data[i]=0;
}
template <int prec>
multi_prec<prec>::multi_prec(const char *s){
  if (read(s, *this)) {
    printf("(qd_real::qd_real): INPUT ERROR.");
    *this = 0.0;
  }
}
template <int prec>
template<int precP>
multi_prec<prec>::multi_prec(const multi_prec<precP> *mp){
	for(int i=0; i<min(prec, precP); i++)	data[i] = mp->getData()[i];
  for(int i=min(prec,precP); i<prec; i++) data[i]=0;
}
template <int prec>
multi_prec<prec>::multi_prec(const multi_prec<prec> *mp){
	for(int i=0; i<prec; i++)	data[i] = mp->getData()[i];
}

//********************* Getters & Setters *********************//
template <int prec>
int multi_prec<prec>::getPrec() const{ return prec; }
template <int prec>
const float* multi_prec<prec>::getData()const{ return data; }
template <int prec>
void multi_prec<prec>::setData(const float *datap, const int precp){  
  for(int i=0; i<min(prec, precp); i++)	data[i]=datap[i];
  for(int i=min(prec, precp); i<prec; i++) data[i]=0;
}
template <int prec>
void multi_prec<prec>::setData(const float *datap){  
  for(int i=0; i<prec; i++)	data[i] = datap[i];
}
template <int prec>
void multi_prec<prec>::setElement(const float datap, const int index){
	data[index] = datap;
}

//********************* Addition *********************//
/** Computes mp1 + mp2 and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |x+y - r| <= **/
template <int pR, int p1, int p2>
void certifAddExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2){
  certifiedAdd<p1,p2,pR>(mp1.data, mp2.data, res.data);
}
template <int pR, int p1>
void certifAddExpans_d(multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val){
  certifiedAdd<p1,1,pR>(mp1.data, &val, res.data);
}

/** Computes mp1 + mp2 and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |x+y - r| <= **/
template <int pR, int p1, int p2>
void truncAddExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2){
  certifiedAdd<(pR<p1)?pR:p1,(pR<p2)?pR:p2,pR>(mp1.data, mp2.data, res.data);
}
template <int pR, int p1>
void truncAddExpans_d(multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val){
  certifiedAdd<(pR<p1)?pR:p1,1,pR>(mp1.data, &val, res.data);
}

/** Uses the generalized QD addition algorithm. It requires an expensive renormalization in order to render the rezult non-overlapping => bad performance **/
template <int pR, int p1, int p2>
void QD_LikeAddExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2){
	float aux[pR+1];
	if(p1>=p2) baileyAdd<p1,p2,pR+1>(mp1.data, mp2.data, aux);
  else baileyAdd<p2,p1,pR+1>(mp2.data, mp1.data, aux);
  fast_renorm2L<pR+1,pR>(aux, res.data);
}
template <int pR, int p1>
void QD_LikeAddExpans_d(multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val){
	float aux[pR+1];
  baileyAdd<p1,1,pR>(mp1.data, &val, aux);
  fast_renorm2L<pR+1,pR>(aux, res.data);
}

//********************* Multiplication *********************//
/** Computes mp1 * mp2 and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |xy - r| <= **/
template <int pR, int p1, int p2>
void certifMulExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2){
	certifiedMul<p1,p2,pR>(mp1.data, mp2.data, res.data);
}
template <int pR, int p1>
void certifMulExpans_d(multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val){
	certifiedMul<p1,1,pR>(mp1.data, &val, res.data);
}

/** Computes mp1 * mp2 and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |xy - r| <= **/
template <int pR, int p1, int p2>
void truncMulExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2){
  if(p1>=p2) truncatedMul<p1,p2,pR>(mp1.data, mp2.data, res.data);
  else truncatedMul<p2,p1,pR>(mp2.data, mp1.data, res.data);
}
template <int pR, int p1>
void truncMulExpans_d(multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val){
  truncatedMul<p1,1,pR>(mp1.data, &val, res.data);
}

/** Uses the generalized QD addition algorithm. It requires an expensive renormalization in order to render the rezult non-overlapping => bad performance **/
template <int pR, int p1, int p2>
void QD_LikeMulExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2){
	float aux[pR+1];
	if(p1>=p2) baileyMul<p1,p2,pR+1>(mp1.data, mp2.data, aux);
  else baileyMul<p2,p1,pR+1>(mp2.data, mp1.data, aux);
  fast_renorm2L<pR+1,pR>(aux, res.data);
}
template <int pR, int p1>
void QD_LikeMulExpans_d(multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val){
	float aux[pR+1];
  baileyMul<p1,1,pR>(mp1.data, &val, aux);
  fast_renorm2L<pR+1,pR>(aux, res.data);
}

//********************* Division *********************//
/** Computes the reciprocal of mp1 and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |1/x - r| <= **/
template <int pR, int p1>
void invExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1){
	invNewtonUp_QD<p1,pR>(mp1.data, res.data);
}
template <int pR>
void invExpans_d(multi_prec<pR> &res, const float val){
	invNewtonUp_QD<1,pR>(&val, res.data);
}

/** Computes mp1 / mp2 and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |x/y - r| <= **/
template <int pR, int p1, int p2>
void divExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1, const multi_prec<p2> &mp2){
  divNewton_QD<p1,p2,pR>(mp1.data, mp2.data, res.data);
}
template <int pR, int p1>
void divExpans_d(multi_prec<pR> &res, const multi_prec<p1> &mp1, const float val){
  divNewton_QD<p1,1,pR>(mp1.data, &val, res.data);
}
template <int pR, int p1>
void divExpans_d(multi_prec<pR> &res, const float val, const multi_prec<p1> &mp1){
  divNewton_QD<1,p1,pR>(&val, mp1.data, res.data);
}

//********************* Square Root *********************//
/** Computes the reciprocal of sqrt(mp1) and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |1/sqrt(x) - r| <= **/
template <int pR, int p1>
void invSqrtExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1){
  sqrtInvNewton_QD<p1,pR>(mp1.data, res.data);
}
template <int pR>
void invSqrtExpans_d(multi_prec<pR> &res, const float val){
  sqrtInvNewton<1,pR>(&val, res.data);
}

/** Computes sqrt(mp1) and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |sqrt(x) - r| <= **/
template <int pR, int p1>
void sqrtNewtonExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1){
  sqrtNewton<p1,pR>(mp1.data, res.data);
}
template <int pR>
void sqrtNewtonExpans_d(multi_prec<pR> &res, const float val){
	sqrtNewton<1,pR>(&val, res.data);
}

/** Computes sqrt(mp1) and puts the result in res (an ulp-nonoverlapping expansion).
    Error bound: |sqrt(x) - r| <= **/
template <int pR, int p1>
void sqrtHeronExpans(multi_prec<pR> &res, const multi_prec<p1> &mp1){
  sqrtHeron<p1,pR>(mp1.data, res.data);
}
template <int pR>
void sqrtHeronExpans_d(multi_prec<pR> &res, const float val){
  sqrtHeron<1,pR>(&val, res.data);
}

/**-----------operators-------------**/
/** Checks equality between to multiprecs.
    If precs are different, the higher precision terms have to be zero **/
template <int p>
template<int pS>
bool multi_prec<p>::operator ==(const multi_prec<pS> &mp2) const{
  for(int i=0; i<min(p,pS); i++)
    if(data[i] != (mp2.getData())[i]) return false;
  if(p<pS)
    for(int i=p;i<pS;i++) if(0 != (mp2.getData())[i]) return false;  
  if(pS<p)
    for(int i=pS;i<p;i++) if(0 != data[i]) return false;  
  return true;
}
template <int p>
bool multi_prec<p>::operator ==(const multi_prec<p> &mp2) const{
	for(int i=0; i<p; i++)
    if(data[i] != (mp2.getData())[i]) return false;  
  return true;
}
template <int p>
bool multi_prec<p>::operator ==(const float &mp2) const{
  if (p==1 && data[0] == mp2) return true;
  else if (data[0] != mp2) return false;
  else if (p>1 && data[0] == mp2)
    for(int i=1; i<p; i++) if(data[i] != 0) return false;
  return true;
}

/** Checks inequality between to multiprecs.
    Works as NOT (operator ==) defined above **/
template <int p>
template <int pS>
bool multi_prec<p>::operator !=(const multi_prec<pS> &mp2) const{ return !((*this) == mp2); }
template <int p>
bool multi_prec<p>::operator !=(const multi_prec<p> &mp2) const{ return !((*this) == mp2); }
template <int p>
bool multi_prec<p>::operator !=(const float &mp2) const{ return !((*this) == mp2); }

/** Checks equality between to multiprecs.
    If precs are different, the higher precision terms have to be zero **/
template <int p>
template<int pS>
bool multi_prec<p>::operator <(const multi_prec<pS> &mp2) const{
  int i;
  for(i=0; i<min(p,pS); i++)
    if(data[i] > mp2.getData()[i]) return false;
    else if(data[i] < mp2.getData()[i]) return true;

  if(p < pS)
    for(; i<pS; i++){
      if(mp2.getData()[i] != 0.)    
        if ((mp2.getData())[i] < 0) return false;
        else return true;
    }
  else if(p > pS)
    for(; i<p; i++){
      if(data[i] < 0) return true;
      else return false;
    }
  return false;
}
template <int p>
bool multi_prec<p>::operator <(const multi_prec<p> &mp2) const{
  for(int i=0; i<p; i++)
    if (data[i] > mp2.getData()[i]) return false;
    else if (data[i] < mp2.getData()[i]) return true;
  return false;
}
template <int p>
bool multi_prec<p>::operator <(const float &mp2) const{
  if(data[0] < mp2) return true;
  else if (data[0] == mp2)
    if (data[1] < 0.) return true;
    else if (data[1] > 0.) return false;
         else
           for(int i=2; i<p; i++)
             if (data[i] != 0.) return false;
  return false;
}

/** Checks equality between to multiprecs.
    If precs are different, the higher precision terms have to be zero **/
template <int p>
template<int pS>
bool multi_prec<p>::operator <=(const multi_prec<pS> &mp2) const{
  return ((*this) < mp2) || ((*this) == mp2);
}
template <int p>
bool multi_prec<p>::operator <=(const multi_prec<p> &mp2) const{
  return ((*this) < mp2) || ((*this) == mp2);
}
template <int p>
bool multi_prec<p>::operator <=(const float &mp2) const{
  return ((*this) < mp2) || ((*this) == mp2);
}

/** Checks equality between to multiprecs.
    If precs are different, the higher precision terms have to be zero **/
template <int p>
template<int pS>
bool multi_prec<p>::operator >(const multi_prec<pS> &mp2) const{
  return !((*this) < mp2) && ((*this) != mp2);
}
template <int p>
bool multi_prec<p>::operator >(const multi_prec<p> &mp2) const{
  return !((*this) < mp2) && ((*this) != mp2);
}
template <int p>
bool multi_prec<p>::operator >(const float &mp2) const{
  return !((*this) < mp2) && ((*this) != mp2);
}

template <int p>
template<int pS>
bool multi_prec<p>::operator >=(const multi_prec<pS> &mp2) const{
  return ((*this) > mp2) || ((*this) == mp2);
}
template <int p>
bool multi_prec<p>::operator >=(const multi_prec<p> &mp2) const{
  return ((*this) > mp2) || ((*this) == mp2);
}
template <int p>
bool multi_prec<p>::operator >=(const float &mp2) const{
  return ((*this) > mp2) || ((*this) == mp2);
}

template <int pR, int p1, int p2>
multi_prec<pR> max( const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 ){
  if (mp1 < mp2) return mp2;
  else return mp1;
}
template <int pR, int p1, int p2>
multi_prec<pR> min( const multi_prec<p1> &mp1, const multi_prec<p2> &mp2 ){
  if (mp1 > mp2) return mp2;
  else return mp1;
}

/** Returns the absolute value of the multi_prec parameter mp
    Remark: if data[0] < 0 returns the negation of mp **/
template <int p>
multi_prec<p> abs(const multi_prec<p> &mp){
 	multi_prec<p> res;
 	res = mp;
	if((res.getData())[0] < 0) return -res;
	return res;
}

template <int prec>
multi_prec<prec> sqrt(const multi_prec<prec> &mp){
	multi_prec<prec> res;
	sqrtNewtonExpans(res, mp);
	return res;
}
template <int prec>
multi_prec<prec> invSqrt(const multi_prec<prec> &mp){
	multi_prec<prec> res;
	invSqrtExpans(res, mp);
	return res;
}

template <int p>
void renorm(multi_prec<p> &mp){ renorm2L<p, p>(mp.data); }
template <int p>
void renorm_rand(multi_prec<p> &mp){ renorm_rand2L<p, p>(mp.data); }
template <int p>
void renorm_2ndL(multi_prec<p> &mp){ fast_VecSumErrBranch<p, p>(mp.data); }

template<int pS>
int read(const char *s, multi_prec<pS> &mp) {
  const char *p = s;
  char ch;
  int sign = 0;
  int point = -1;  /* location of decimal point */
  int nd = 0;      /* number of digits read */
  int e = 0;       /* exponent. */
  bool done = false;
  multi_prec<pS> r = 0.0;  /* number being read */

  /* Skip any leading spaces */
  while (*p == ' ') p++;
  while (!done && (ch = *p) != '\0') {
    if (ch >= '0' && ch <= '9') { /* It's a digit */
      int d = ch - '0';
      r *= 10.0;
      r += static_cast<float>(d);
      nd++;
    } else { /* Non-digit */
      switch (ch) {
      case '.':
        if (point >= 0) return -1;   /* we've already encountered a decimal point. */
        point = nd;
        break;
      case '-':
      case '+':
        if (sign != 0 || nd > 0) return -1;  /* we've already encountered a sign, or if its
                                                not at first position. */
        sign = (ch == '-') ? -1 : 1;
        break;
      case 'E':
      case 'e':
        int nread;
        nread = std::sscanf(p+1, "%d", &e);
        done = true;
        if (nread != 1) return -1;  /* read of exponent failed. */
        break;
      case ' ':
        done = true;
        break;
      default:
        return -1;
      }
    }
    p++;
  }

  /* Adjust exponent to account for decimal point */
  if (point >= 0) e -= (nd - point);

  /* Multiply the the exponent */
  if (e != 0) r *= pow(10,e); //(qd_real(10.0) ^ e);

  mp = (sign < 0) ? -r : r;
  return 0;
}

#endif
