/*
 * specAddition.h
 *
 * This file is part of CAMPARY Library
 *
 * Copyright (C) 2014 - 
 *
 * CAMPARY Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CAMPARY Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
 
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */

#ifndef _specAddition_h
#define _specAddition_h

/****************************************************/
template <>
__host__ __device__ __forceinline__ void merge<1,1>(float const *x, float const *y, float *z){
  if(fabs(y[0])>fabs(x[0])){ z[0] = y[0]; z[1] = x[0]; } else { z[0] = x[0]; z[1] = y[0]; }
}

/****************************************************/
template <>
__host__ __device__ __forceinline__ void certifiedAdd<1,1,1>(const float *x, const float *y, float *r){ r[0] = FPadd_rn(x[0], y[0]); }
template <>
__host__ __device__ __forceinline__ void certifiedAdd<1,1,2>(const float *x, const float *y, float *r){ r[0] = two_sum(x[0], y[0], r[1]); }
template <>
__host__ __device__ __forceinline__ void certifiedAdd<1,2,2>(const float *x, const float *y, float *r){ 
	float f[3];
	merge<1,2>( x, y, f );

	f[1] = fast_two_sum(f[1], f[2], f[2]);
	r[0] = two_sum(f[0], f[1], f[1]);
	if(f[1] == 0.) r[0] = fast_two_sum(r[0], f[2], r[1]);
	else r[1] = f[1] + f[2];
}
template <>
__host__ __device__ __forceinline__ void certifiedAdd<2,1,2>(const float *x, const float *y, float *r){ 
	float f[3];
	merge<2,1>( x, y, f );

	f[1] = fast_two_sum(f[1], f[2], f[2]);
	r[0] = two_sum(f[0], f[1], f[1]);
	if(f[1] == 0.) r[0] = fast_two_sum(r[0], f[2], r[1]);
	else r[1] = f[1] + f[2];
}

/****************************************************/
template <>
__host__ __device__ __forceinline__ void baileyAdd<1,1,1>(const float *x, const float *y, float *z){ z[0] = FPadd_rn(x[0], y[0]); }
template <>
__host__ __device__ __forceinline__ void baileyAdd<1,1,2>(const float *x, const float *y, float *z){ z[0] = two_sum(x[0], y[0], z[1]); }
template <>
__host__ __device__ __forceinline__ void baileyAdd<2,1,3>(const float *x, const float *y, float *z){ 
	z[0] = two_sum( x[0], y[0], z[1] );
	z[1] = two_sum( z[1], x[1], z[2] );
}
template<>
__host__ __device__ __forceinline__ void baileyAdd<2,2,3>(float const *x, float const *y, float *z){
	float e;
  z[1] = two_sum ( x[1], y[1], z[2] );
  z[0] = two_sum ( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e ); 
  z[2] = FPadd_rn( z[2], e );
}
template<>
__host__ __device__ __forceinline__ void baileyAdd<3,1,4>(float const *x, float const *y, float *z){
  float e;
  z[0] = two_sum( x[0], y[0], e );
  z[1] = two_sum( x[1], e,    e );
  z[2] = two_sum( x[2], e,    z[3] );
}
template<>
__host__ __device__ __forceinline__ void baileyAdd<3,3,4>(float const *x, float const *y, float *z){
	float e;
  z[2] = two_sum ( x[2], y[2], z[3] );
  z[1] = two_sum ( x[1], y[1], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
  z[0] = two_sum ( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
}
template<>
__host__ __device__ __forceinline__ void baileyAdd<4,1,5>(float const *x, float const *y, float *z){
  float e;
  z[0] = two_sum( x[0], y[0], e );
  z[1] = two_sum( x[1], e, e );
  z[2] = two_sum( x[2], e, e );
  z[3] = two_sum( x[3], e, z[4] );
}
template<>
__host__ __device__ __forceinline__ void baileyAdd<4,4,5>(float const *x, float const *y, float *z){
  float e;
  z[3] = two_sum ( x[3], y[3], z[4] );
  z[2] = two_sum ( x[2], y[2], e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[1] = two_sum ( x[1], y[1], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[0] = two_sum ( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
}

template<>
__host__ __device__ __forceinline__ void baileyAdd<3,2,3>(float const *x, float const *y, float *z){
  float e;
  z[1] = two_sum ( x[1], y[1], e );
  z[2] = FPadd_rn( x[2], e );
  z[0] = two_sum ( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = FPadd_rn( z[2], e );
}
template<>
__host__ __device__ __forceinline__ void baileyAdd<3,2,4>(float const *x, float const *y, float *z){
  float e;
  z[1] = two_sum ( x[1], y[1], e );
  z[2] = two_sum ( x[2], e,    z[3] );
  z[0] = two_sum ( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
}

template<>
__host__ __device__ __forceinline__ void baileyAdd<4,2,3>(float const *x, float const *y, float *z){
  float e;
  z[1] = two_sum ( x[1], y[1], e );
  z[2] = FPadd_rn( x[2], e );
  z[0] = two_sum ( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = FPadd_rn( z[2], e );
}
template<>
__host__ __device__ __forceinline__ void baileyAdd<4,2,5>(float const *x, float const *y, float *z){
  float e;
  z[1] = two_sum ( x[1], y[1], e );
  z[2] = two_sum ( x[2], e,    e );
  z[3] = two_sum ( x[3], e,    z[4] );
  z[0] = two_sum ( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
}

#endif
