/*
 * specMultiplication.h
 *
 * This file is part of CAMPARY Library
 *
 * Copyright (C) 2014 - 
 *
 * CAMPARY Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CAMPARY Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
 
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */

#ifndef _specMultiplication_h
#define _specMultiplication_h

/****************************************************/
template<> __host__ __device__ __forceinline__ void certifiedMul<1,1,1>(float const *x, float const *y, float *r){ r[0] = FPmul_rn(x[0], y[0]); }
template<> __host__ __device__ __forceinline__ void certifiedMul<1,1,2>(float const *x, float const *y, float *r){ r[0] = two_prod(x[0], y[0], r[1]); }

/****************************************************/
template<> __host__ __device__ __forceinline__ void truncatedMul<1,1,1>(float const *x, float const *y, float *z){ z[0] = FPmul_rn(x[0], y[0]); }
template<> __host__ __device__ __forceinline__ void truncatedMul<1,1,2>(float const *x, float const *y, float *z){ z[0] = two_prod(x[0], y[0], z[1]); }

template<>
__host__ __device__ __forceinline__ void truncatedMul<2,1,2>(float const *x, float const *y, float *z){
  float e, f[3];
  f[1] = two_prod( x[1], y[0], f[2] );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e ); 
  f[2] = FPadd_rn( f[2], e );
	renorm_rand2L<3,2>(f, z);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<2,2,2>(float const *x, float const *y, float *z){
  float p, e, f[3];
  f[2] = FPmul_rn( x[1], y[1] );
  f[1] = two_prod( x[0], y[1], e );
  f[2] = FPadd_rn( f[2], e );
     p = two_prod( x[1], y[0], e );
  f[2] = FPadd_rn( f[2], e );
  f[1] = two_sum ( f[1], p,    e );
  f[2] = FPadd_rn( f[2], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = FPadd_rn( f[2], e );
	renorm_rand2L<3,2>(f, z);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<3,1,3>(float const *x, float const *y, float *z){
  float e, f[4];
  f[2] = two_prod( x[2], y[0], f[3] );
  f[1] = two_prod( x[1], y[0], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
	renorm_rand2L<4,3>(f, z);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<3,3,3>(float const *x, float const *y, float *z){
  float p, e, f[4];
  f[3] = FPmul_rn( x[1], y[2] );
  f[3] = FPadd_rn( f[3], FPmul_rn( x[2], y[1] ) );
  f[2] = two_prod( x[0], y[2], e );
  f[3] = FPadd_rn( f[3], e );
     p = two_prod( x[1], y[1], e );
  f[3] = FPadd_rn( f[3], e );
  f[2] = two_sum ( f[2], p,    e );
  f[3] = FPadd_rn( f[3], e );
     p = two_prod( x[2], y[0], e );
  f[3] = FPadd_rn( f[3], e );
  f[2] = two_sum ( f[2], p,    e );
  f[3] = FPadd_rn( f[3], e );
  f[1] = two_prod( x[0], y[1], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
     p = two_prod( x[1], y[0], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
  f[1] = two_sum ( f[1], p,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
  renorm_rand2L<4,3>(f, z);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<4,1,4>(float const *x, float const *y, float *z){
  float e, f[5];
  f[3] = two_prod( x[3], y[0], f[4] );
  f[2] = two_prod( x[2], y[0], e );
  f[3] = two_sum ( f[3], e,    e ); 
  f[4] = FPadd_rn( f[4], e );
  f[1] = two_prod( x[1], y[0], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  renorm_rand2L<5,4>(f, z);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<4,4,4>(float const *x, float const *y, float *z){
  float p, e, f[5];
  f[4] = FPmul_rn( x[1], y[3] );
  f[4] = FPadd_rn( f[4], FPmul_rn( x[2], y[2] ) );
  f[4] = FPadd_rn( f[4], FPmul_rn( x[3], y[1] ) );
  f[3] = two_prod( x[0], y[3], e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[1], y[2], e );
  f[4] = FPadd_rn( f[4], e );
  f[3] = two_sum ( f[3], p,    e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[2], y[1], e );
  f[4] = FPadd_rn( f[4], e );
  f[3] = two_sum ( f[3], p,    e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[3], y[0], e );
  f[4] = FPadd_rn( f[4], e );
  f[3] = two_sum ( f[3], p,    e );
  f[4] = FPadd_rn( f[4], e );
  f[2] = two_prod( x[0], y[2], e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[1], y[1], e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[2] = two_sum ( f[2], p,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[2], y[0], e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[2] = two_sum ( f[2], p,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[1] = two_prod( x[0], y[1], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[1], y[0], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[1] = two_sum ( f[1], p,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
	renorm_rand2L<5,4>(f, z);
}

template<>
__host__ __device__ __forceinline__ void truncatedMul<3,2,2>(float const *x, float const *y, float *z){
  float p, e, f[3];
  f[2] = FPmul_rn( x[1], y[1] );
  f[2] = FPadd_rn( f[2], FPmul_rn( x[2], y[0] ));
  f[1] = two_prod( x[0], y[1], e );
  f[2] = FPadd_rn( f[2], e );
     p = two_prod( x[1], y[0], e );
  f[2] = FPadd_rn( f[2], e );
  f[1] = two_sum ( f[1], p,    e );
  f[2] = FPadd_rn( f[2], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = FPadd_rn( f[2], e );
  renorm_rand2L<3,2>(f, z);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<3,2,3>(float const *x, float const *y, float *z){
  float p, e, f[4];
  f[3] = FPmul_rn( x[2], y[1] );
  
  f[2] = two_prod( x[2], y[0], e );
  f[3] = FPadd_rn( f[3], e );
     p = two_prod( x[1], y[1], e ); 
  f[3] = FPadd_rn( f[3], e );
  f[2] = two_sum ( f[2], p,    e );
  f[3] = FPadd_rn( f[3], e );
  
  f[1] = two_prod( x[0], y[1], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
     p = two_prod( x[1], y[0], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
  f[1] = two_sum ( f[1], p,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = FPadd_rn( f[3], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = two_sum ( f[2], e,   e );
  f[3] = FPadd_rn( f[3], e );
  renorm_rand2L<4,3>(f, z);
}

template<>
__host__ __device__ __forceinline__ void truncatedMul<4,2,2>(float const *x, float const *y, float *z){
  float p, e, f[3];
  f[2] = FPmul_rn( x[1], y[1] );
  f[2] = FPadd_rn( f[2], FPmul_rn( x[2], y[0] ));
  f[1] = two_prod( x[0], y[1], e );
  f[2] = FPadd_rn( f[2], e );
     p = two_prod( x[1], y[0], e );
  f[2] = FPadd_rn( f[2], e );
  f[1] = two_sum ( f[1], p,    e );
  f[2] = FPadd_rn( f[2], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = FPadd_rn( f[2], e );
  renorm_rand2L<3,2>(f, z);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<4,2,4>(float const *x, float const *y, float *z){
  float p, e, f[5];
  f[4] = FPmul_rn( x[3], y[1] );
  f[3] = two_prod( x[3], y[0], e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[2], y[1], e );
  f[4] = FPadd_rn( f[4], e );
  f[3] = two_sum ( f[3], p,    e );
  f[4] = FPadd_rn( f[4], e );
  f[2] = two_prod( x[2], y[0], e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[1], y[1], e );
  f[3] = two_sum ( f[3], e,    e ); 
  f[4] = FPadd_rn( f[4], e );
  f[2] = two_sum ( f[2], p,    e );
  f[3] = two_sum ( f[3], e,    e ); 
  f[4] = FPadd_rn( f[4], e );
  f[1] = two_prod( x[0], y[1], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
     p = two_prod( x[1], y[0], e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[1] = two_sum ( f[1], p,    e );
  f[2] = two_sum ( f[2], e,    e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  f[0] = two_prod( x[0], y[0], e );
  f[1] = two_sum ( f[1], e,    e );
  f[2] = two_sum ( f[2], e,   e );
  f[3] = two_sum ( f[3], e,    e );
  f[4] = FPadd_rn( f[4], e );
  renorm_rand2L<5,4>(f, z);
}

/****************************************************/
template<> __host__ __device__ __forceinline__ void baileyMul<1,1,1>(float const *x, float const *y, float *z){ z[0] = FPmul_rn(x[0], y[0]); }
template<> __host__ __device__ __forceinline__ void baileyMul<1,1,2>(float const *x, float const *y, float *z){ z[0] = two_prod(x[0], y[0], z[1]); }
template<>
__host__ __device__ __forceinline__ void baileyMul<2,1,3>(float const *x, float const *y, float *z){
  float e;
  z[1] = two_prod( x[1], y[0], z[2] );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e ); 
  z[2] = FPadd_rn( z[2], e );
}
template<>
__host__ __device__ __forceinline__ void baileyMul<2,2,3>(float const *x, float const *y, float *z){
  float p, e;
  z[2] = FPmul_rn( x[1], y[1] );
  z[1] = two_prod( x[0], y[1], e );
  z[2] = FPadd_rn( z[2], e );
     p = two_prod( x[1], y[0], e );
  z[2] = FPadd_rn( z[2], e );
  z[1] = two_sum ( z[1], p,    e );
  z[2] = FPadd_rn( z[2], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = FPadd_rn( z[2], e );
}
template<>
__host__ __device__ __forceinline__ void baileyMul<3,1,4>(float const *x, float const *y, float *z){
  float e;
  z[2] = two_prod( x[2], y[0], z[3] );
  z[1] = two_prod( x[1], y[0], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
}
template<>
__host__ __device__ __forceinline__ void baileyMul<3,3,4>(float const *x, float const *y, float *z){
  float p, e;
  z[3] = FPmul_rn( x[1], y[2] );
  z[3] = FPadd_rn( z[3], FPmul_rn( x[2], y[1] ) );
  z[2] = two_prod( x[0], y[2], e );
  z[3] = FPadd_rn( z[3], e );
     p = two_prod( x[1], y[1], e );
  z[3] = FPadd_rn( z[3], e );
  z[2] = two_sum ( z[2], p,    e );
  z[3] = FPadd_rn( z[3], e );
     p = two_prod( x[2], y[0], e );
  z[3] = FPadd_rn( z[3], e );
  z[2] = two_sum ( z[2], p,    e );
  z[3] = FPadd_rn( z[3], e );
  z[1] = two_prod( x[0], y[1], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
     p = two_prod( x[1], y[0], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
  z[1] = two_sum ( z[1], p,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
}
template<>
__host__ __device__ __forceinline__ void baileyMul<4,1,5>(float const *x, float const *y, float *z){
  float e;
  z[3] = two_prod( x[3], y[0], z[4] );
  z[2] = two_prod( x[2], y[0], e );
  z[3] = two_sum ( z[3], e,    e ); 
  z[4] = FPadd_rn( z[4], e );
  z[1] = two_prod( x[1], y[0], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
}
template<>
__host__ __device__ __forceinline__ void baileyMul<4,4,5>(float const *x, float const *y, float *z){
  float p, e;
  z[4] = FPmul_rn( x[1], y[3] );
  z[4] = FPadd_rn( z[4], FPmul_rn( x[2], y[2] ) );
  z[4] = FPadd_rn( z[4], FPmul_rn( x[3], y[1] ) );
  z[3] = two_prod( x[0], y[3], e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[1], y[2], e );
  z[4] = FPadd_rn( z[4], e );
  z[3] = two_sum ( z[3], p,    e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[2], y[1], e );
  z[4] = FPadd_rn( z[4], e );
  z[3] = two_sum ( z[3], p,    e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[3], y[0], e );
  z[4] = FPadd_rn( z[4], e );
  z[3] = two_sum ( z[3], p,    e );
  z[4] = FPadd_rn( z[4], e );
  z[2] = two_prod( x[0], y[2], e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[1], y[1], e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[2] = two_sum ( z[2], p,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[2], y[0], e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[2] = two_sum ( z[2], p,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[1] = two_prod( x[0], y[1], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[1], y[0], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[1] = two_sum ( z[1], p,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
}

template<>
__host__ __device__ __forceinline__ void baileyMul<3,2,3>(float const *x, float const *y, float *z){
  float p, e;
  z[2] = FPmul_rn( x[1], y[1] );
  z[2] = FPadd_rn( z[2], FPmul_rn( x[2], y[0] ));
  z[1] = two_prod( x[0], y[1], e );
  z[2] = FPadd_rn( z[2], e );
     p = two_prod( x[1], y[0], e );
  z[2] = FPadd_rn( z[2], e );
  z[1] = two_sum ( z[1], p,    e );
  z[2] = FPadd_rn( z[2], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = FPadd_rn( z[2], e );
}
template<>
__host__ __device__ __forceinline__ void baileyMul<3,2,4>(float const *x, float const *y, float *z){
  float p, e;
  z[3] = FPmul_rn( x[2], y[1] );
  
  z[2] = two_prod( x[2], y[0], e );
  z[3] = FPadd_rn( z[3], e );
     p = two_prod( x[1], y[1], e ); 
  z[3] = FPadd_rn( z[3], e );
  z[2] = two_sum ( z[2], p,    e );
  z[3] = FPadd_rn( z[3], e );
  
  z[1] = two_prod( x[0], y[1], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
     p = two_prod( x[1], y[0], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
  z[1] = two_sum ( z[1], p,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = FPadd_rn( z[3], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,   e );
  z[3] = FPadd_rn( z[3], e );
}

template<>
__host__ __device__ __forceinline__ void baileyMul<4,2,3>(float const *x, float const *y, float *z){
  float p, e;
  z[2] = FPmul_rn( x[1], y[1] );
  z[2] = FPadd_rn( z[2], FPmul_rn( x[2], y[0] ));
  z[1] = two_prod( x[0], y[1], e );
  z[2] = FPadd_rn( z[2], e );
     p = two_prod( x[1], y[0], e );
  z[2] = FPadd_rn( z[2], e );
  z[1] = two_sum ( z[1], p,    e );
  z[2] = FPadd_rn( z[2], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = FPadd_rn( z[2], e );
}
template<>
__host__ __device__ __forceinline__ void baileyMul<4,2,5>(float const *x, float const *y, float *z){
  float p, e;
  z[4] = FPmul_rn( x[3], y[1] );
  z[3] = two_prod( x[3], y[0], e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[2], y[1], e );
  z[4] = FPadd_rn( z[4], e );
  z[3] = two_sum ( z[3], p,    e );
  z[4] = FPadd_rn( z[4], e );
  z[2] = two_prod( x[2], y[0], e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[1], y[1], e );
  z[3] = two_sum ( z[3], e,    e ); 
  z[4] = FPadd_rn( z[4], e );
  z[2] = two_sum ( z[2], p,    e );
  z[3] = two_sum ( z[3], e,    e ); 
  z[4] = FPadd_rn( z[4], e );
  z[1] = two_prod( x[0], y[1], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
     p = two_prod( x[1], y[0], e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[1] = two_sum ( z[1], p,    e );
  z[2] = two_sum ( z[2], e,    e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
  z[0] = two_prod( x[0], y[0], e );
  z[1] = two_sum ( z[1], e,    e );
  z[2] = two_sum ( z[2], e,   e );
  z[3] = two_sum ( z[3], e,    e );
  z[4] = FPadd_rn( z[4], e );
}

#endif
