/*
 * specRenorm.h
 *
 * This file is part of CAMPARY Library
 *
 * Copyright (C) 2014 - 
 *
 * CAMPARY Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CAMPARY Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
 
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */

#ifndef _specRenorm_h
#define _specRenorm_h

/****************************************************/
template<> __host__ __device__ __forceinline__ void fast_VecSum<1>(float *x){ }
template<> __host__ __device__ __forceinline__ void fast_VecSum<1>(const float *x, float *r){ r[0] = x[0]; }
template<> __host__ __device__ __forceinline__ void fast_VecSum<2>(float *x){ x[0] = fast_two_sum(x[0], x[1], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_VecSum<2>(const float *x, float *r){ r[0] = fast_two_sum(x[0], x[1], r[1]); }

/****************************************************/
template<> __host__ __device__ __forceinline__ void VecSum<1>(float *x){ }
template<> __host__ __device__ __forceinline__ void VecSum<1>(const float *x, float *r){ r[0] = x[0]; }
template<> __host__ __device__ __forceinline__ void VecSum<2>(float *x){ x[0] = two_sum(x[0], x[1], x[1]); }
template<> __host__ __device__ __forceinline__ void VecSum<2>(const float *x, float *r){ r[0] = two_sum(x[0], x[1], r[1]); }

/****************************************************/
template<> __host__ __device__ __forceinline__ void fast_VecSumErrBranch<1,1>(float *x){ }
template<> __host__ __device__ __forceinline__ void fast_VecSumErrBranch<1,1>(const float *x, float *r){ r[0] = x[0]; }
template<> __host__ __device__ __forceinline__ void fast_VecSumErrBranch<2,1>(float *x){ x[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_VecSumErrBranch<2,1>(const float *x, float *r){ r[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_VecSumErrBranch<2,2>(float *x){ x[0] = fast_two_sum(x[0], x[1], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_VecSumErrBranch<2,2>(const float *x, float *r){ r[0] = fast_two_sum(x[0], x[1], r[1]); }

/****************************************************/
template<> __host__ __device__ __forceinline__ void fast_VecSumErr<1>(float *x){ }
template<> __host__ __device__ __forceinline__ void fast_VecSumErr<1>(const float *x, float *r){ r[0] = x[0]; }
template<> __host__ __device__ __forceinline__ void fast_VecSumErr<2>(float *x){ x[0] = fast_two_sum(x[0], x[1], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_VecSumErr<2>(const float *x, float *r){ r[0] = fast_two_sum(x[0], x[1], r[1]); }

/****************************************************/
template<> __host__ __device__ __forceinline__ void fast_renorm2L<1,1>(float *x){ }
template<> __host__ __device__ __forceinline__ void fast_renorm2L<1,1>(const float *x, float *r){ r[0] = x[0]; }
template<> __host__ __device__ __forceinline__ void fast_renorm2L<2,1>(float *x){ x[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_renorm2L<2,1>(const float *x, float *r){ r[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_renorm2L<2,2>(float *x){ x[0] = fast_two_sum(x[0], x[1], x[1]); }
template<> __host__ __device__ __forceinline__ void fast_renorm2L<2,2>(const float *x, float *r){ r[0] = fast_two_sum(x[0], x[1], r[1]); }

/****************************************************/
template<> __host__ __device__ __forceinline__ void renorm2L<1,1>(float *x){ }
template<> __host__ __device__ __forceinline__ void renorm2L<1,1>(const float *x, float *r){ r[0] = x[0]; }
template<> __host__ __device__ __forceinline__ void renorm2L<2,1>(float *x){ x[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void renorm2L<2,1>(const float *x, float *r){ r[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void renorm2L<2,2>(float *x){ x[0] = two_sum(x[0], x[1], x[1]); }
template<> __host__ __device__ __forceinline__ void renorm2L<2,2>(const float *x, float *r){ r[0] = two_sum(x[0], x[1], r[1]); }

/****************************************************/
template<> __host__ __device__ __forceinline__ void renorm_rand2L<1,1>(float *x){ }
template<> __host__ __device__ __forceinline__ void renorm_rand2L<1,1>(const float *x, float *r){ r[0] = x[0]; }
template<> __host__ __device__ __forceinline__ void renorm_rand2L<2,1>(float *x){ x[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void renorm_rand2L<2,1>(const float *x, float *r){ r[0] = FPadd_rn(x[0], x[1]); }
template<> __host__ __device__ __forceinline__ void renorm_rand2L<2,2>(float *x){ x[0] = two_sum(x[0], x[1], x[1]); }
template<> __host__ __device__ __forceinline__ void renorm_rand2L<2,2>(const float *x, float *r){ r[0] = two_sum(x[0], x[1], r[1]); }

#endif
